@@ -9,100 +9,149 @@ use Carp;
 use List::Util qw(first);
 use WebService::Linode::Base;
 
-our $VERSION = '0.11';
+our $VERSION = '0.12';
 our @ISA     = ("WebService::Linode::Base");
 our $AUTOLOAD;
 
 my %validation = (
     account => {
-        info => [ [], [] ],
-    },
-    api => {
-        spec => [ [], [] ],
+        estimateinvoice => [ ['mode'], [qw( paymentterm linodeid planid )] ],
+        info            => [ [],       [] ],
+        paybalance      => [ [],       [] ],
+        updatecard => [ [qw( ccexpmonth ccexpyear ccnumber )], [] ],
     },
+    api => { spec => [ [], [] ], },
     avail => {
-        datacenters => [ [], [] ],
-        distributions => [ [], [ 'distributionid' ] ],
-        kernels => [ [], [ 'isxen', 'kernelid' ] ],
-        linodeplans => [ [ 'planid' ], [] ],
-        stackscripts => [ [], [qw( keywords distributionid distributionvendor )] ],
+        datacenters   => [ [], [] ],
+        distributions => [ [], ['distributionid'] ],
+        kernels       => [ [], [ 'kernelid', 'isxen' ] ],
+        linodeplans   => [ [], ['planid'] ],
+        stackscripts =>
+            [ [], [qw( distributionid keywords distributionvendor )] ],
     },
     domain => {
-        create => [ [ 'domain', 'type' ], [qw( status ttl_sec expire_sec master_ips refresh_sec soa_email retry_sec axfr_ips description )] ],
-        delete => [ [ 'domainid' ], [] ],
-        list => [ [], [ 'domainid' ] ],
-        update => [ [ 'domainid' ], [qw( status domain ttl_sec expire_sec type master_ips refresh_sec soa_email axfr_ips retry_sec description )] ],
+        create => [
+            [ 'type', 'domain' ],
+            [   qw( refresh_sec retry_sec master_ips expire_sec soa_email axfr_ips description ttl_sec status )
+            ]
+        ],
+        delete => [ ['domainid'], [] ],
+        list   => [ [],           ['domainid'] ],
+        update => [
+            ['domainid'],
+            [   qw( refresh_sec retry_sec master_ips type expire_sec domain soa_email axfr_ips description ttl_sec status )
+            ]
+        ],
     },
     domain_resource => {
-        create => [ [qw( domainid type )], [qw( target ttl_sec port weight priority protocol name )] ],
+        create => [
+            [qw( type domainid )],
+            [qw( protocol name weight target priority ttl_sec port )]
+        ],
         delete => [ [ 'resourceid', 'domainid' ], [] ],
-        list => [ [ 'domainid' ], [ 'resourceid' ] ],
-        update => [ [ 'resourceid' ], [qw( target domainid ttl_sec port weight priority protocol name )] ],
+        list => [ ['domainid'], ['resourceid'] ],
+        update => [
+            ['resourceid'],
+            [qw( weight target priority ttl_sec domainid port protocol name )]
+        ],
     },
     linode => {
-        boot => [ [ 'linodeid' ], [ 'configid' ] ],
-        clone => [ [qw( planid paymentterm linodeid datacenterid )], [] ],
-        create => [ [qw( planid paymentterm datacenterid )], [] ],
-        delete => [ [ 'linodeid' ], [ 'skipchecks' ] ],
-        list => [ [], [ 'linodeid' ] ],
-        reboot => [ [ 'linodeid' ], [ 'configid' ] ],
-        resize => [ [ 'planid', 'linodeid' ], [] ],
-        shutdown => [ [ 'linodeid' ], [] ],
-        update => [ [ 'linodeid' ], [qw( alert_bwquota_threshold alert_bwin_threshold alert_cpu_threshold alert_cpu_enabled alert_diskio_enabled label backupweeklyday alert_bwquota_enabled watchdog lpm_displaygroup alert_bwin_enabled alert_bwout_enabled alert_bwout_threshold alert_diskio_threshold backupwindow )] ],
+        boot => [ ['linodeid'], ['configid'] ],
+        clone => [ [qw( linodeid paymentterm datacenterid planid )], [] ],
+        create => [ [qw( datacenterid planid paymentterm )], [] ],
+        delete => [ ['linodeid'], ['skipchecks'] ],
+        list   => [ [],           ['linodeid'] ],
+        mutate => [ ['linodeid'], [] ],
+        reboot => [ ['linodeid'], ['configid'] ],
+        resize => [ [ 'linodeid', 'planid' ], [] ],
+        shutdown => [ ['linodeid'], [] ],
+        update => [
+            ['linodeid'],
+            [   qw( alert_diskio_threshold lpm_displaygroup watchdog alert_bwout_threshold ms_ssh_disabled ms_ssh_ip ms_ssh_user alert_bwout_enabled alert_diskio_enabled ms_ssh_port alert_bwquota_enabled alert_bwin_threshold backupweeklyday alert_cpu_enabled alert_bwquota_threshold backupwindow alert_cpu_threshold alert_bwin_enabled label )
+            ]
+        ],
+        webconsoletoken => [ ['linodeid'], [] ],
     },
     linode_config => {
-        create => [ [qw( linodeid label kernelid )], [qw( comments helper_xen devtmpfs_automount rootdevicecustom rootdevicero helper_depmod helper_disableupdatedb rootdevicenum disklist runlevel ramlimit )] ],
-        delete => [ [ 'configid', 'linodeid' ], [] ],
-        list => [ [ 'linodeid' ], [ 'configid' ] ],
-        update => [ [ 'configid' ], [qw( comments helper_xen devtmpfs_automount rootdevicecustom linodeid rootdevicero label helper_depmod helper_disableupdatedb rootdevicenum disklist runlevel kernelid ramlimit )] ],
+        create => [
+            [qw( kernelid linodeid label )],
+            [   qw( rootdevicero helper_disableupdatedb rootdevicenum comments rootdevicecustom devtmpfs_automount ramlimit runlevel helper_depmod helper_xen disklist )
+            ]
+        ],
+        delete => [ [ 'linodeid', 'configid' ], [] ],
+        list => [ ['linodeid'], ['configid'] ],
+        update => [
+            ['configid'],
+            [   qw( helper_disableupdatedb rootdevicero comments rootdevicenum rootdevicecustom kernelid runlevel ramlimit devtmpfs_automount helper_depmod linodeid helper_xen disklist label )
+            ]
+        ],
     },
     linode_disk => {
-        create => [ [qw( type size linodeid label )], [] ],
-        createfromdistribution => [ [qw( size rootpass linodeid distributionid label )], [ 'rootsshkey' ] ],
-        createfromstackscript => [ [qw( size linodeid rootpass distributionid stackscriptudfresponses stackscriptid label )], [] ],
-        delete => [ [ 'diskid', 'linodeid' ], [] ],
-        duplicate => [ [ 'diskid', 'linodeid' ], [] ],
-        list => [ [ 'linodeid' ], [ 'diskid' ] ],
+        create => [ [qw( label size type linodeid )], [] ],
+        createfromdistribution => [
+            [qw( rootpass linodeid distributionid size label )],
+            ['rootsshkey']
+        ],
+        createfromstackscript => [
+            [   qw( size label linodeid stackscriptid distributionid rootpass stackscriptudfresponses )
+            ],
+            []
+        ],
+        delete    => [ [ 'linodeid', 'diskid' ],   [] ],
+        duplicate => [ [ 'diskid',   'linodeid' ], [] ],
+        list => [ ['linodeid'], ['diskid'] ],
         resize => [ [qw( diskid linodeid size )], [] ],
-        update => [ [ 'diskid' ], [qw( linodeid isreadonly label )] ],
+        update => [ ['diskid'], [qw( linodeid label isreadonly )] ],
     },
     linode_ip => {
-        addprivate => [ [ 'linodeid' ], [] ],
-        list => [ [ 'linodeid' ], [ 'ipaddressid' ] ],
-    },
-    linode_job => {
-        list => [ [ 'linodeid' ], [ 'pendingonly', 'jobid' ] ],
+        addprivate => [ ['linodeid'], [] ],
+        list       => [ ['linodeid'], ['ipaddressid'] ],
     },
+    linode_job => { list => [ ['linodeid'], [ 'pendingonly', 'jobid' ] ], },
     nodebalancer => {
-        create => [ [ 'paymentterm', 'datacenterid' ], [ 'label', 'clientconnthrottle' ] ],
-        delete => [ [ 'nodebalancerid' ], [] ],
-        list => [ [], [ 'nodebalancerid' ] ],
-        update => [ [ 'nodebalancerid' ], [ 'label', 'clientconnthrottle' ] ],
+        create => [
+            [ 'paymentterm',        'datacenterid' ],
+            [ 'clientconnthrottle', 'label' ]
+        ],
+        delete => [ ['nodebalancerid'], [] ],
+        list   => [ [],                 ['nodebalancerid'] ],
+        update => [ ['nodebalancerid'], [ 'label', 'clientconnthrottle' ] ],
     },
     nodebalancer_config => {
-        create => [ [ 'nodebalancerid' ], [qw( check_path check_body stickiness port check check_timeout check_attempts check_interval protocol algorithm )] ],
-        delete => [ [ 'configid' ], [] ],
-        list => [ [ 'nodebalancerid' ], [ 'configid' ] ],
-        update => [ [ 'configid' ], [qw( check_path check_body stickiness port check_timeout check check_attempts check_interval protocol algorithm )] ],
+        create => [
+            ['nodebalancerid'],
+            [   qw( protocol check check_path check_interval algorithm check_attempts stickiness check_timeout check_body port )
+            ]
+        ],
+        delete => [ ['configid'],       [] ],
+        list   => [ ['nodebalancerid'], ['configid'] ],
+        update => [
+            ['configid'],
+            [   qw( check_body stickiness check_attempts check_timeout algorithm port check protocol check_path check_interval )
+            ]
+        ],
     },
     nodebalancer_node => {
-        create => [ [qw( configid address label )], [ 'mode', 'weight' ] ],
-        delete => [ [ 'nodeid' ], [] ],
-        list => [ [ 'configid' ], [ 'nodeid' ] ],
-        update => [ [ 'nodeid' ], [qw( address mode weight label )] ],
+        create => [ [qw( label address configid )], [ 'mode', 'weight' ] ],
+        delete => [ ['nodeid'],                     [] ],
+        list   => [ ['configid'],                   ['nodeid'] ],
+        update => [ ['nodeid'], [qw( mode label address weight )] ],
     },
     stackscript => {
-        create => [ [qw( script distributionidlist label )], [qw( rev_note description ispublic )] ],
-        delete => [ [ 'stackscriptid' ], [] ],
-        list => [ [], [ 'stackscriptid' ] ],
-        update => [ [ 'stackscriptid' ], [qw( script rev_note ispublic label distributionidlist description )] ],
-    },
-    test => {
-        echo => [ [], [] ],
-    },
-    user => {
-        getapikey => [ [ 'password', 'username' ], [] ],
+        create => [
+            [qw( label distributionidlist script )],
+            [qw( rev_note description ispublic )]
+        ],
+        delete => [ ['stackscriptid'], [] ],
+        list   => [ [],                ['stackscriptid'] ],
+        update => [
+            ['stackscriptid'],
+            [   qw( distributionidlist description script ispublic rev_note label )
+            ]
+        ],
     },
+    test => { echo => [ [], [] ], },
+    user => { getapikey => [ [ 'username', 'password' ], [] ], },
 );
 
 sub AUTOLOAD {
@@ -210,10 +259,10 @@ Optional Parameters:
 
 =over 4
 
-=item * keywords
-
 =item * distributionid
 
+=item * keywords
+
 =item * distributionvendor
 
 =back
@@ -230,10 +279,10 @@ Optional Parameters:
 
 =over 4
 
-=item * isxen
-
 =item * kernelid
 
+=item * isxen
+
 =back
 
 =head3 avail_linodeplans
@@ -242,14 +291,14 @@ Required Parameters:
 
 =over 4
 
-=item * planid
-
 =back
 
 Optional Parameters:
 
 =over 4
 
+=item * planid
+
 =back
 
 =head3 avail_datacenters
@@ -288,34 +337,34 @@ Required Parameters:
 
 =over 4
 
-=item * domain
-
 =item * type
 
+=item * domain
+
 =back
 
 Optional Parameters:
 
 =over 4
 
-=item * status
+=item * refresh_sec
 
-=item * ttl_sec
-
-=item * expire_sec
+=item * retry_sec
 
 =item * master_ips
 
-=item * refresh_sec
+=item * expire_sec
 
 =item * soa_email
 
-=item * retry_sec
-
 =item * axfr_ips
 
 =item * description
 
+=item * ttl_sec
+
+=item * status
+
 =back
 
 =head3 domain_delete
@@ -348,28 +397,28 @@ Optional Parameters:
 
 =over 4
 
-=item * status
+=item * refresh_sec
 
-=item * domain
-
-=item * ttl_sec
-
-=item * expire_sec
-
-=item * type
+=item * retry_sec
 
 =item * master_ips
 
-=item * refresh_sec
+=item * type
+
+=item * expire_sec
+
+=item * domain
 
 =item * soa_email
 
 =item * axfr_ips
 
-=item * retry_sec
-
 =item * description
 
+=item * ttl_sec
+
+=item * status
+
 =back
 
 =head3 domain_list
@@ -394,30 +443,30 @@ Required Parameters:
 
 =over 4
 
-=item * domainid
-
 =item * type
 
+=item * domainid
+
 =back
 
 Optional Parameters:
 
 =over 4
 
+=item * protocol
+
+=item * name
+
+=item * weight
+
 =item * target
 
+=item * priority
+
 =item * ttl_sec
 
 =item * port
 
-=item * weight
-
-=item * priority
-
-=item * protocol
-
-=item * name
-
 =back
 
 =head3 domain_resource_delete
@@ -452,18 +501,18 @@ Optional Parameters:
 
 =over 4
 
+=item * weight
+
 =item * target
 
-=item * domainid
+=item * priority
 
 =item * ttl_sec
 
+=item * domainid
+
 =item * port
 
-=item * weight
-
-=item * priority
-
 =item * protocol
 
 =item * name
@@ -488,17 +537,13 @@ Optional Parameters:
 
 =back
 
-=head3 linode_create
+=head3 linode_mutate
 
 Required Parameters:
 
 =over 4
 
-=item * planid
-
-=item * paymentterm
-
-=item * datacenterid
+=item * linodeid
 
 =back
 
@@ -508,15 +553,17 @@ Optional Parameters:
 
 =back
 
-=head3 linode_resize
+=head3 linode_create
 
 Required Parameters:
 
 =over 4
 
+=item * datacenterid
+
 =item * planid
 
-=item * linodeid
+=item * paymentterm
 
 =back
 
@@ -544,6 +591,22 @@ Optional Parameters:
 
 =back
 
+=head3 linode_webconsoletoken
+
+Required Parameters:
+
+=over 4
+
+=item * linodeid
+
+=back
+
+Optional Parameters:
+
+=over 4
+
+=back
+
 =head3 linode_boot
 
 Required Parameters:
@@ -562,20 +625,38 @@ Optional Parameters:
 
 =back
 
+=head3 linode_resize
+
+Required Parameters:
+
+=over 4
+
+=item * linodeid
+
+=item * planid
+
+=back
+
+Optional Parameters:
+
+=over 4
+
+=back
+
 =head3 linode_clone
 
 Required Parameters:
 
 =over 4
 
-=item * planid
+=item * linodeid
 
 =item * paymentterm
 
-=item * linodeid
-
 =item * datacenterid
 
+=item * planid
+
 =back
 
 Optional Parameters:
@@ -632,36 +713,44 @@ Optional Parameters:
 
 =over 4
 
-=item * alert_bwquota_threshold
-
-=item * alert_bwin_threshold
-
-=item * alert_cpu_threshold
-
-=item * alert_cpu_enabled
-
-=item * alert_diskio_enabled
-
-=item * label
-
-=item * backupweeklyday
-
-=item * alert_bwquota_enabled
-
-=item * watchdog
+=item * alert_diskio_threshold
 
 =item * lpm_displaygroup
 
-=item * alert_bwin_enabled
-
-=item * alert_bwout_enabled
+=item * watchdog
 
 =item * alert_bwout_threshold
 
-=item * alert_diskio_threshold
+=item * ms_ssh_disabled
+
+=item * ms_ssh_ip
+
+=item * ms_ssh_user
+
+=item * alert_bwout_enabled
+
+=item * alert_diskio_enabled
+
+=item * ms_ssh_port
+
+=item * alert_bwquota_enabled
+
+=item * alert_bwin_threshold
+
+=item * backupweeklyday
+
+=item * alert_cpu_enabled
+
+=item * alert_bwquota_threshold
 
 =item * backupwindow
 
+=item * alert_cpu_threshold
+
+=item * alert_bwin_enabled
+
+=item * label
+
 =back
 
 =head3 linode_list
@@ -686,39 +775,39 @@ Required Parameters:
 
 =over 4
 
+=item * kernelid
+
 =item * linodeid
 
 =item * label
 
-=item * kernelid
-
 =back
 
 Optional Parameters:
 
 =over 4
 
-=item * comments
-
-=item * helper_xen
-
-=item * devtmpfs_automount
-
-=item * rootdevicecustom
-
 =item * rootdevicero
 
-=item * helper_depmod
-
 =item * helper_disableupdatedb
 
 =item * rootdevicenum
 
-=item * disklist
+=item * comments
+
+=item * rootdevicecustom
+
+=item * devtmpfs_automount
+
+=item * ramlimit
 
 =item * runlevel
 
-=item * ramlimit
+=item * helper_depmod
+
+=item * helper_xen
+
+=item * disklist
 
 =back
 
@@ -728,10 +817,10 @@ Required Parameters:
 
 =over 4
 
-=item * configid
-
 =item * linodeid
 
+=item * configid
+
 =back
 
 Optional Parameters:
@@ -754,34 +843,34 @@ Optional Parameters:
 
 =over 4
 
-=item * comments
-
-=item * helper_xen
-
-=item * devtmpfs_automount
-
-=item * rootdevicecustom
-
-=item * linodeid
+=item * helper_disableupdatedb
 
 =item * rootdevicero
 
-=item * label
-
-=item * helper_depmod
-
-=item * helper_disableupdatedb
+=item * comments
 
 =item * rootdevicenum
 
-=item * disklist
-
-=item * runlevel
+=item * rootdevicecustom
 
 =item * kernelid
 
+=item * runlevel
+
 =item * ramlimit
 
+=item * devtmpfs_automount
+
+=item * helper_depmod
+
+=item * linodeid
+
+=item * helper_xen
+
+=item * disklist
+
+=item * label
+
 =back
 
 =head3 linode_config_list
@@ -808,13 +897,13 @@ Required Parameters:
 
 =over 4
 
-=item * type
+=item * label
 
 =item * size
 
-=item * linodeid
+=item * type
 
-=item * label
+=item * linodeid
 
 =back
 
@@ -850,14 +939,14 @@ Required Parameters:
 
 =over 4
 
-=item * size
-
 =item * rootpass
 
 =item * linodeid
 
 =item * distributionid
 
+=item * size
+
 =item * label
 
 =back
@@ -894,10 +983,10 @@ Required Parameters:
 
 =over 4
 
-=item * diskid
-
 =item * linodeid
 
+=item * diskid
+
 =back
 
 Optional Parameters:
@@ -922,10 +1011,10 @@ Optional Parameters:
 
 =item * linodeid
 
-=item * isreadonly
-
 =item * label
 
+=item * isreadonly
+
 =back
 
 =head3 linode_disk_list
@@ -954,17 +1043,17 @@ Required Parameters:
 
 =item * size
 
+=item * label
+
 =item * linodeid
 
-=item * rootpass
-
-=item * distributionid
-
-=item * stackscriptudfresponses
-
 =item * stackscriptid
 
-=item * label
+=item * distributionid
+
+=item * rootpass
+
+=item * stackscriptudfresponses
 
 =back
 
@@ -1034,11 +1123,11 @@ Required Parameters:
 
 =over 4
 
-=item * script
+=item * label
 
 =item * distributionidlist
 
-=item * label
+=item * script
 
 =back
 
@@ -1084,18 +1173,18 @@ Optional Parameters:
 
 =over 4
 
-=item * script
-
-=item * rev_note
-
-=item * ispublic
-
-=item * label
-
 =item * distributionidlist
 
 =item * description
 
+=item * script
+
+=item * ispublic
+
+=item * rev_note
+
+=item * label
+
 =back
 
 =head3 stackscript_list
@@ -1128,26 +1217,26 @@ Optional Parameters:
 
 =over 4
 
-=item * check_path
-
-=item * check_body
-
-=item * stickiness
-
-=item * port
+=item * protocol
 
 =item * check
 
-=item * check_timeout
-
-=item * check_attempts
+=item * check_path
 
 =item * check_interval
 
-=item * protocol
-
 =item * algorithm
 
+=item * check_attempts
+
+=item * stickiness
+
+=item * check_timeout
+
+=item * check_body
+
+=item * port
+
 =back
 
 =head3 nodebalancer_config_delete
@@ -1180,25 +1269,25 @@ Optional Parameters:
 
 =over 4
 
-=item * check_path
-
 =item * check_body
 
 =item * stickiness
 
-=item * port
+=item * check_attempts
 
 =item * check_timeout
 
+=item * algorithm
+
+=item * port
+
 =item * check
 
-=item * check_attempts
-
-=item * check_interval
-
 =item * protocol
 
-=item * algorithm
+=item * check_path
+
+=item * check_interval
 
 =back
 
@@ -1226,11 +1315,11 @@ Required Parameters:
 
 =over 4
 
-=item * configid
+=item * label
 
 =item * address
 
-=item * label
+=item * configid
 
 =back
 
@@ -1274,14 +1363,14 @@ Optional Parameters:
 
 =over 4
 
-=item * address
-
 =item * mode
 
-=item * weight
-
 =item * label
 
+=item * address
+
+=item * weight
+
 =back
 
 =head3 nodebalancer_node_list
@@ -1308,10 +1397,10 @@ Required Parameters:
 
 =over 4
 
-=item * password
-
 =item * username
 
+=item * password
+
 =back
 
 Optional Parameters:
@@ -1322,6 +1411,8 @@ Optional Parameters:
 
 =head1 Additional Helper Methods
 
+These methods are deprecated and will be going away.
+
 =head3 getDomainIDbyName( domain => 'example.com' )
 
 Returns the ID for a domain given the name.
@@ -1340,49 +1431,6 @@ Takes a record name and domainid or domain and returns the resourceid.
 
 =back
 
-=head1 BUGS
-
-This module does not yet support the Linode API batch method, patches welcome.
-
-Please report any bugs or feature requests to C<bug-webservice-linode
-at rt.cpan.org>, or through the web interface at
-L<http://rt.cpan.org/NoAuth/ReportBug.html?Queue=WebService-Linode>.  I will
-be notified, and then you'll automatically be notified of progress on your
-bug as I make changes.
-
-=head1 SUPPORT
-
-You can find documentation for this module with the perldoc command.
-
-    perldoc WebService::Linode
-
-
-You can also look for information at:
-
-=over 4
-
-=item * Module Repo
-
-L<http://git.thegrebs.com/?p=WebService-Linode;a=summary>
-
-=item * RT: CPAN's request tracker
-
-L<http://rt.cpan.org/NoAuth/Bugs.html?Dist=WebService-Linode>
-
-=item * AnnoCPAN: Annotated CPAN documentation
-
-L<http://annocpan.org/dist/WebService-Linode>
-
-=item * CPAN Ratings
-
-L<http://cpanratings.perl.org/d/WebService-Linode>
-
-=item * Search CPAN
-
-L<http://search.cpan.org/dist/WebService-Linode>
-
-=back
-
 =head1 COPYRIGHT & LICENSE
 
 Copyright 2008-2009 Linode, LLC, all rights reserved.
